# OVirtDataWrapper.pm
# interface for accessing oVirt data:
#   provides filepaths

package OVirtDataWrapper;

use strict;

use JSON;
use Data::Dumper;
use Xorux_lib qw(error read_json);

defined $ENV{INPUTDIR} || Xorux_lib::error( "INPUTDIR undefined, config etc/lpar2rrd.cfg probably has not been loaded " . __FILE__ . ":" . __LINE__ ) && exit 1;

my $inputdir        = $ENV{INPUTDIR};
my $wrkdir          = "$inputdir/data";
my $dc_path         = "$wrkdir/oVirt";
my $host_path       = "$wrkdir/oVirt/host";
my $storage_path    = "$wrkdir/oVirt/storage";
my $vms_path        = "$wrkdir/oVirt/vm";
my $metadata_file   = "$wrkdir/oVirt/metadata.json";
my %dictionary      = ();
my $metadata_loaded = 0;

################################################################################

sub get_filepath_rrd_host {
  my $host_uuid = shift;

  return "$host_path/$host_uuid/sys.rrd";
}

sub get_filepath_rrd_host_nic {
  my $host_uuid = shift;
  my $nic_uuid  = shift;

  return "$host_path/$host_uuid/nic-$nic_uuid.rrd";
}

sub get_filepath_rrd_disk {
  my $disk_uuid = shift;

  return "$storage_path/disk-$disk_uuid.rrd";
}

sub get_filepath_rrd_vm {
  my $vm_uuid = shift;

  return "$vms_path/$vm_uuid/sys.rrd";
}

sub get_filepath_rrd_vm_nic {
  my $vm_uuid  = shift;
  my $nic_uuid = shift;

  return "$vms_path/$vm_uuid/nic-$nic_uuid.rrd";
}

sub get_filepath_rrd_storage_domain {
  my $storage_domain_uuid = shift;

  return "$storage_path/sd-$storage_domain_uuid.rrd";
}

sub get_filepath_rrd {
  my $type = shift;

  if    ( $type eq 'host' )           { return get_filepath_rrd_host( @_ ); }
  elsif ( $type eq 'host_nic' )       { return get_filepath_rrd_host_nic( @_ ); }
  elsif ( $type eq 'disk' )           { return get_filepath_rrd_disk( @_ ); }
  elsif ( $type eq 'vm' )             { return get_filepath_rrd_vm( @_ ); }
  elsif ( $type eq 'vm_nic' )         { return get_filepath_rrd_vm_nic( @_ ); }
  elsif ( $type eq 'storage_domain' ) { return get_filepath_rrd_storage_domain( @_ ); }
  else {
    Xorux_lib::error( "Unknown rrd type $type: " . __FILE__ . ":" . __LINE__ );
    return '';
  }
}

################################################################################

sub get_conf {
  if ( ! $metadata_loaded && -f $metadata_file ) {
    my ( $code, $ref ) = Xorux_lib::read_json( $metadata_file );
    $metadata_loaded = $code;
    %dictionary = $code ? %{ $ref } : ();
  }

  return \%dictionary;
} ## end sub get_conf

sub get_conf_section {
  my $section    = shift;
  my $dictionary = get_conf();
  my $ref;

  if ( $section eq "labels" ) {
    $ref = $dictionary->{labels};
  } elsif ( $section eq "mapping" ) {
    $ref = $dictionary->{mapping};
  } elsif ( $section eq "arch" ) {
    $ref = $dictionary->{architecture};
  } elsif ( $section eq "arch-dc" ) {
    $ref = $dictionary->{architecture}{datacenter};
  } elsif ( $section eq "arch-cl" ) {
    $ref = $dictionary->{architecture}{cluster};
  } elsif ( $section eq "arch-host" ) {
    $ref = $dictionary->{architecture}{host};
  } elsif ( $section eq "arch-sd" ) {
    $ref = $dictionary->{architecture}{storage_domain};
  } elsif ( $section eq "arch-vm" ) {
    $ref = $dictionary->{architecture}{vm};
  } else {
    $ref = {};
  }

  return defined $ref && ref( $ref ) eq 'HASH' ? $ref : {};
} ## end sub get_conf_section

# get_arch ( $uuid,
#            $type,
#            $subtype
#          )
#  return: [ $uuid1, $uuid2  ... ] i.e. reference to list of child uuids of subtype entries
#
sub get_arch {
  my $uuid    = shift;
  my $type    = shift;
  my $subtype = shift;
  my $conf    = get_conf();

  return ( exists $conf->{architecture}{ $type }{ $uuid }{ $subtype } )
    ? $conf->{architecture}{ $type }{ $uuid }{ $subtype } : [];
} ## end sub get_arch

# get_parent ( $type,
#              $parent_type,
#              $uuid )
#  return: parent_uuid
sub get_parent {
  my $type    = shift;
  my $uuid    = shift;
  my $conf    = get_conf();

  return $conf->{architecture}{ $type }{ $uuid }{parent};
} ## end sub get_parent

# get_labels ( )
#  return: ( type1 => { uuid1 => "label1", ... }, type2 => { ... }, ... )
sub get_labels {
  my $dictionary = get_conf();

  return $dictionary->{labels};
} ## end sub get_labels

# get_label ( $type,
#              $uuid )
#  return: label
sub get_label {
  my $type       = shift;
  my $uuid       = shift;
  my $dictionary = get_conf();

  return ( exists $dictionary->{labels}{ $type }{ $uuid } ) ? $dictionary->{labels}{ $type }{ $uuid } : $uuid;
} ## end get_label

# get_mapping ( $uuid )
#  return: agent_dirname
sub get_mapping {
  my $uuid       = shift;
  my $dictionary = get_conf();

  return ( exists $dictionary->{mapping}{ $uuid } ) ? $dictionary->{mapping}{ $uuid } : undef;
} ## end sub get_mapping

# get_uuids ( $type )
#  return: [ $uuid1, $uuid2,  ... ]
sub get_uuids {
  my $type = shift;
  my $conf = get_conf();
  my $ref  = [];

  if ( exists $conf->{labels}{ $type } ) {
    my @keys = keys %{ $conf->{labels}{ $type } };
    $ref = \@keys;
  }

  return $ref;
} ## end sub get_uuids

################################################################################

1;
