# oVirtMenu.pm
# page types and associated tools for generating front-end menu and tabs for OVirtServer

package OVirtMenu;

use strict;

use JSON;
use Data::Dumper;

my @page_types = (
                  {
                   type       => 'vm',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'vm' ],
                   tabs       => [
                                  { cpu_core => 'CPU' },
                                  { cpu_percent => 'CPU %' },
                                  { mem => 'MEM' },
                                  { aggr_net => 'Net' },
                                  { aggr_data => 'Data' },
                                  { aggr_latency => 'Latency' }
                                 ]
                  },
                  {
                   type       => 'host',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'host' ],
                   tabs       => [
                                  { cpu_core => 'CPU' },
                                  { cpu_percent => 'CPU %' },
                                  { mem => 'MEM' },
                                  { aggr_vm_cpu_core => 'VM CPU' },
                                  { aggr_vm_mem_used => 'VM MEM used' },
                                  { aggr_vm_mem_free => 'VM MEM free' }
                                 ]
                  },
                  {
                   type       => 'host_nic',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'host', 'nic' ],
                   tabs       => [
                                  { net => 'Net' }
                                 ]
                  },
                  {
                   type       => 'host_nic_aggr',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'host' ],
                   tabs       => [
                                  { net => 'Net' }
                                 ]
                  },
                  {
                   type       => 'storage_domain',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'storage_domain' ],
                   tabs       => [
                                  { space => 'Space' },
                                  { aggr_data => 'Data' },
                                  { aggr_latency => 'Latency' }
                                 ]
                  },
                  {
                   type       => 'disk',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'disk' ],
                   tabs       => [
                                  { data => 'Data' },
                                  { latency => 'Latency' }
                                 ]
                  },
                  {
                   type       => 'disk_aggr',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'storage_domain' ],
                   tabs       => [
                                  { data => 'Data' },
                                  { latency => 'Latency' }
                                 ]
                  },
                  {
                   type       => 'cluster_aggr',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [ 'cluster' ],
                   tabs       => [
                                  { host_cpu_core => 'CPU' },
                                  { host_cpu_percent => 'CPU %' },
                                  { host_mem_used => 'MEM used' },
                                  { host_mem_free => 'MEM free' },
                                  { host_mem_used_percent => 'MEM used %' },
                                  { vm_cpu_core => 'VM CPU'},
                                  { vm_cpu_percent => 'VM CPU %'},
                                  { vm_mem_used => 'VM MEM used'},
                                  { vm_mem_free => 'VM MEM free'},
                                  { vm_mem_used_percent => 'VM MEM used %'}
                                 ]
                  },
                  {
                   type       => 'configuration',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [],
                   tabs       => [
                                  { host_conf => 'Host' },
                                  { vm_conf => 'VM' },
                                  { storage_domain_conf => 'Storage Domain' }
                                 ]
                  },
                  {
                   type       => 'storage_domains_total_aggr',
                   platform   => 'oVirt',
                   url_base   => '/lpar2rrd-cgi/detail.sh',
                   url_params => [],
                   tabs       => [
                                  { data => 'Data' },
                                  { latency => 'Latency' }
                                 ]
                  },
                  {
                   type       => 'Heatmap',
                   platform   => 'oVirt',
                   url_base   => 'heatmap-ovirt.html',
                   url_params => [],
                   tabs       => []
                  }
                 );

sub create_folder {
  my $title  = shift;
  my $str    = shift;
  my %folder = ( "folder" => "true", "title" => $title, children => [] );

  $folder{str} = $str if defined $str;

  return \%folder;
}

sub create_page {
  my $title = shift;
  my $url   = shift;
  my $str   = shift;
  my $agent = shift;

  my %page  = ( "title" => $title, "str" => defined $str ? $str : $title, "href" => $url );

  if ( $title eq 'Heatmap' ) {
    $page{extraClasses} = 'boldmenu';
  }

  if ( defined $agent ) {
    $page{agent} = $agent;
  }

  return \%page;
}

sub get_url {
  # accepts a hash with named parameters: type, and respective url_params
  my ( $args ) = @_;
  my $url;

  foreach my $page_type ( @page_types ) {
    if ( $page_type->{type} eq $args->{type} ) {
      $url = $page_type->{url_base} =~ /\.html$/ ? $page_type->{url_base}
        : "$page_type->{url_base}?platform=$page_type->{platform}&type=$page_type->{type}";

      foreach my $param ( @{ $page_type->{url_params} } ) {
        $url .= "&$param=$args->{$param}";
      }
      last;
    }
  }

  return $url;
}

sub get_tabs {
  my $type = shift;

  for my $page_type (@page_types) {
    if ( $page_type->{type} eq $type ) {
      return $page_type->{tabs};
    }
  }
}

1;
