# XenServerDataWrapper.pm
# interface for accessing XenServer data:
#   provides filepaths, lists objects with RRDs, accesses metadata in `conf.json`

package XenServerDataWrapper;

use strict;

use JSON;
use Data::Dumper;
use Xorux_lib qw(error);

defined $ENV{INPUTDIR} || Xorux_lib::error( " INPUTDIR undefined, config etc/lpar2rrd.cfg probably has not been loaded " . __FILE__ . ":" . __LINE__ ) && exit 1;

my $inputdir = $ENV{INPUTDIR};
my $wrkdir   = "$inputdir/data";

my $hosts_path  = "$wrkdir/XEN";
my $vms_path    = "$wrkdir/XEN_VMs";
my $conf_file   = "$wrkdir/XEN_iostats/conf.json";

################################################################################

sub get_filepath_rrd_host {
  my $uuid     = shift;
  my $filepath = "$hosts_path/$uuid/sys.rrd";

  return $filepath;
}

sub get_filepath_rrd_host_lan {
  my $uuid     = shift;
  my $id       = shift;
  my $filepath = "$hosts_path/$uuid/lan-$id.rrd";

  return $filepath;
}

sub get_filepath_rrd_host_disk {
  my $uuid     = shift;
  my $id       = shift;
  my $filepath = "$hosts_path/$uuid/disk-$id.rrd";

  return $filepath;
}

sub get_filepath_rrd_vm {
  my $uuid     = shift;
  my $filepath = "$vms_path/$uuid.rrd";

  return $filepath;
}

################################################################################

sub get_host_list {
  my @hosts = ();

  # get all hosts' directories
  opendir( DIR, "$hosts_path" ) || Xorux_lib::error( " cannot open directory : $hosts_path " . __FILE__ . ":" . __LINE__ ) && return 1;
  my @hosts_dir = grep !/^\.\.?$/, readdir(DIR);
  closedir(DIR);

  foreach my $host (@hosts_dir) {
    unless ( -d "$hosts_path/$host" ) { next; }
    push @hosts, $host;
  }

  return \@hosts;
}

sub get_host_in_pool_list {
  my $pool = shift;

  my @hosts = ();
  my $pools = XenServerDataWrapper::get_conf_section("arch-pool");
  my @reported_hosts = ( $pools && exists $pools->{$pool} ) ? @{ $pools->{$pool} } : ();
  foreach my $host (@reported_hosts) {
    if ( -f get_filepath_rrd_host($host) ) {
      push @hosts, $host;
    }
  }

  return \@hosts;
}

sub get_vm_in_pool_list {
  my $pool = shift;

  my @pool_vms = ();
  my $host_vm = XenServerDataWrapper::get_conf_section("arch-host-vm");
  my $hosts = XenServerDataWrapper::get_host_in_pool_list($pool);
  foreach my $host (@{$hosts}) {
    my @host_vms = ( $host_vm && exists $host_vm->{$host} ) ? @{ $host_vm->{$host} } : ();
    push @pool_vms, @host_vms;
  }

  my %vms;
  @vms{@pool_vms} = ();
  my @result = keys %vms;
  return \@result;
}

sub get_vm_under_host_list {
  my $host = shift;

  my $host_vm = XenServerDataWrapper::get_conf_section("arch-host-vm");
  my @vms = ( $host_vm && exists $host_vm->{$host} ) ? @{ $host_vm->{$host} } : ();

  return \@vms;
}

# get a list of host's storage or network interfaces
sub get_host_interface_list {
  my $host = shift;
  my $type = shift; # "storage" or "lan"

  my @result = ();
  if ( -d "$hosts_path/$host" ) {
    opendir( DIR, "$hosts_path/$host" ) || Xorux_lib::error( " cannot open directory : $hosts_path/$host " . __FILE__ . ":" . __LINE__ ) && return 1;
    my @host_dir = grep !/^\.\.?$/, readdir(DIR);
    closedir(DIR);

    my $regex = qr/\.rrd$/;
    my @filtered_files = grep /$regex/, @host_dir;

    foreach my $file (@filtered_files) {
      if ( $file =~ m/^disk-(.*)\.rrd$/ && $type eq "storage" ) {
        my $id = $1;
        push @result, $id;
      } elsif ( $file =~ m/^lan-(.*)\.rrd$/ && $type eq "lan" ) {
        my $id = $1;
        push @result, $id;
      }
    }
  }

  return \@result;
}

sub get_vm_list {
  my @vms = ();

  # get all VM RRDs' filenames
  opendir( DIR, "$vms_path" ) || Xorux_lib::error( " cannot open directory : $vms_path " . __FILE__ . ":" . __LINE__ ) && return 1;
  my @vm_dir = grep !/^\.\.?$/, readdir(DIR);
  closedir(DIR);

  my $regex = qr/\.rrd$/;
  my @filtered_files = grep /$regex/, @vm_dir;

  foreach my $file (@vm_dir) {
    unless ( -f "$vms_path/$file" ) { next; }

    my $uuid = $file;
    $uuid =~ s/$regex//;

    push @vms, $uuid;
  }

  return \@vms;
}

################################################################################

sub get_conf {
  my %dictionary = ();
  {
    my $content;
    local $/;
    if ( open( FH, "< $conf_file" ) ) {
      $content = <FH>;
      close(FH);
      %dictionary = %{decode_json( $content )};
    }
  }
  return \%dictionary;
}

sub get_conf_section {
  my $section = shift;
  my $dictionary = get_conf();

  if ( $section eq "labels" ) {
    return $dictionary->{labels};
  } elsif ( $section eq "arch" ) {
    return $dictionary->{architecture};
  } elsif ( $section eq "arch-host-vm" ) {
    return $dictionary->{architecture}{host_vm};
  } elsif ( $section eq "arch-pool" ) {
    return $dictionary->{architecture}{pool};
  } elsif ( $section eq "arch-storage" ) {
    return $dictionary->{architecture}{storage};
  } elsif ( $section eq "spec-host" ) {
    return $dictionary->{specification}{host};
  } elsif ( $section eq "spec-vm" ) {
    return $dictionary->{specification}{vm};
  } elsif ( $section eq "spec-sr" ) {
    return $dictionary->{specification}{sr};
  } elsif ( $section eq "spec-vdi" ) {
    return $dictionary->{specification}{vdi};
  } else {
    return ();
  }
}

sub get_dom_parent {
  my $dom_type = shift; # value: 'host' or 'vm'

  # TODO
}

sub complete_sr_uuid {
  my $sr_uuid_start = shift;
  my $host          = shift;
  my $dictionary = get_conf_section("arch-storage");

  foreach my $sr (keys %{$dictionary->{sr_host}}) {
    if ( $sr =~ m/^$sr_uuid_start/ ) {
      foreach my $sr_host (@{$dictionary->{sr_host}{$sr}}) {
        if ( $sr_host eq $host ) {
          return $sr;
        }
      }
    }
  }

  return 0;
}

sub get_labels {
  return get_conf_section("labels");
}

sub get_label {
  my $type  = shift; # "pool", "host", "vm", "sr", "vdi"
  my $uuid  = shift;
  my $labels = get_labels();

  if ( $type eq "network" ) { return $uuid; }

  return ( exists $labels->{$type}{$uuid} ) ? $labels->{$type}{$uuid} : $uuid;
}

sub get_host_cpu_count {
  my $uuid       = shift;
  my $dictionary = get_conf_section("spec-host");

  return ( exists $dictionary->{$uuid}{cpu_count} ) ?  $dictionary->{$uuid}{cpu_count} : -1;
}

sub get_conf_update_time {
  return ( stat( $conf_file ) )[9];
}

1;
