# XenServerMenu.pm
# page types and associated tools for generating front-end menu and tabs for XenServer

package XenServerMenu;

use strict;

use JSON;
use Data::Dumper;

# expected menu structure
#
# folder:XenServer
#   folder:Pool1
#     page:Totals; url:platform=XenServer&type=pool-aggr&pool=UUID
#     folder:host=Host1
#       page:System; url:platform=XenServer&type=host&pool=UUID&host=UUID
#       folder:LAN
#         page:Totals; url:platform=XenServer&type=net-aggr&pool=UUID&host=UUID
#         folder:Items
#           page:eth0; url:platform=XenServer&type=net&pool=UUID&host=UUID&net=ID
#       folder:Storage
#         page:Totals; url:platform=XenServer&type=storage-aggr&pool=UUID&host=UUID
#         folder:Items
#           page:Storage1; url:platform=XenServer&type=storage&pool=UUID&host=UUID&storage=UUID
#   folder:VM
#     page:VirtualMachine1; url:platform=XenServer&type=vm&vm=UUID

my @page_types = (
                  {
                   type       => "pool-aggr",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "pool" ],
                   tabs       => [
                                  { cpu_cores => "CPU" },
                                  { cpu_percent => "CPU %" },
                                  { vm_cpu_cores => "VM CPU" },
                                  { vm_cpu_percent => "VM CPU %" },
                                  { memory_used => "MEM used" },
                                  { memory_free => "MEM free" },
                                  { memory_used => "VM MEM used" },
                                  { memory_free => "VM MEM free" }
                                 ]
                  },
                  {
                   type       => "pool-storage-aggr",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "pool" ],
                   tabs       => [
                                  { storage => "Data" },
                                  { iops => "IOPS" },
                                  { latency => "Latency" }
                                 ]
                  },
                  {
                   type       => "host",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "host" ],
                   tabs       => [
                                  { cpu_cores => "CPU" },
                                  { cpu_percent => "CPU %" },
                                  { vm_cpu_cores => "VM CPU" },
                                  { vm_cpu_percent => "VM CPU %" },
                                  { memory => "MEM" },
                                  { memory_used => "VM MEM used" },
                                  { memory_free => "VM MEM free" }
                                 ]
                  },
                  {
                   type       => "net-aggr",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "host" ],
                   tabs       => [
                                  { net => "Net" }
                                 ]
                  },
                  {
                   type       => "net",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "host", "net" ],
                   tabs       => [
                                  { net => "Net" }
                                 ]
                  },
                  {
                   type       => "storage-aggr",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "host" ],
                   tabs       => [
                                  { storage => "Data" },
                                  { iops => "IOPS" },
                                  { latency => "Latency" }
                                 ]
                  },
                  {
                   type       => "storage",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "host", "storage" ],
                   tabs       => [
                                  { storage => "Data" },
                                  { iops => "IOPS" },
                                  { latency => "Latency" }
                                 ]
                  },
                  {
                   type       => "vm",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [ "vm" ],
                   tabs       => [
                                  { cpu_cores => "CPU" },
                                  { cpu_percent => "CPU %" },
                                  { memory => "MEM" },
                                  { storage => "Data" },
                                  { iops => "IOPS" },
                                  { latency => "Latency" },
                                  { net => "Net" }
                                 ]
                  },
                  {
                   type       => "configuration",
                   platform   => "XenServer",
                   url_base   => "/lpar2rrd-cgi/detail.sh",
                   url_params => [],
                   tabs       => [
                                  { hosts => "Host" },
                                  { vms => "VM" },
                                  { storage => "Storage" },
                                  { storage_vm => "Storage-VM" }
                                 ]
                  },
                  {
                   type       => "Heatmap",
                   platform   => "XenServer",
                   url_base   => "heatmap-xen.html",
                   url_params => [],
                   tabs       => []
                  }
                 );

sub create_folder {
  my $title = shift;
  my %folder = ( "folder" => "true", "title" => $title, children => [] );

  return \%folder;
}

sub create_page {
  my $title = shift;
  my $url = shift;
  my %page = ( "title" => $title, "str" => $title, "href" => $url );

  if ( $title eq 'Heatmap' ) {
    $page{extraClasses} = 'boldmenu';
  }

  return \%page;
}

sub get_url {
  # accepts a hash with named parameters: type, and respective url_params
  my ($args) = @_;

  my $url;
  foreach my $page_type (@page_types) {
    if ( $page_type->{type} eq $args->{type} ) {
      $url = $page_type->{url_base} =~ /\.html$/ ? $page_type->{url_base}
        : "$page_type->{url_base}?platform=$page_type->{platform}&type=$page_type->{type}";

      foreach my $param (@{ $page_type->{url_params} }) {
        $url .= "&$param=$args->{$param}";
      }
      last;
    }
  }

  return $url;
}

sub get_tabs {
  my $type = shift;

  for my $page_type (@page_types) {
    if ( $page_type->{type} eq $type ) {
      return $page_type->{tabs};
    }
  }
}

1;
