# xen-genmenu.pl
# generate menu tree from XenServer RRDs and save it in a JSON file
# replacement for an older equivalent to `find_active_lpar.pl` that generated `menu.txt`

use 5.008_008;

use strict;
use warnings;

use JSON;
use Data::Dumper;

use XenServerDataWrapper;
use XenServerMenu;
use Xorux_lib qw(error);

defined $ENV{INPUTDIR} || error( " INPUTDIR undefined, config etc/lpar2rrd.cfg probably has not been loaded " . __FILE__ . ":" . __LINE__ ) && exit 1;

my $inputdir = $ENV{INPUTDIR};
my $wrkdir   = "$inputdir/data";

################################################################################

unless ( -d "$wrkdir/XEN" || -d "$wrkdir/XEN_VMs" ) {
  exit;
}

my $menu_tree = XenServerMenu::create_folder( "XenServer" );

# add Heatmap page
my $heatmap_page_url = XenServerMenu::get_url( { type => 'Heatmap' } );
my %heatmap_page     = %{ XenServerMenu::create_page( 'Heatmap', $heatmap_page_url ) };
push @{$menu_tree->{children}}, \%heatmap_page;

# add Configuration page
my $config_page_url = XenServerMenu::get_url( { type => 'configuration' } );
my %config_page     = %{ XenServerMenu::create_page( 'Configuration', $config_page_url ) };
push @{$menu_tree->{children}}, \%config_page;

# flag that indicates, if any host exists, thus if the menu should be created at all
my $host_exists = 0;

# add pools with hosts
if ( -d "$wrkdir/XEN" ) {
  # iterate over existing pools
  my $pools = XenServerDataWrapper::get_conf_section("arch-pool");
  foreach my $pool (keys %{$pools}) {
    my ( $host_count, @host_entries, $label, $pool_folder, $totals_url );

    # get an array of hosts in the pool
    my $hosts = XenServerDataWrapper::get_host_in_pool_list($pool);

    # skip empty pools
    $host_count  = scalar @$hosts;
    unless ( $host_count > 0 ) { next; }
    $host_exists = 1;

    # keep track of storages in the whole pool
    my $storage_total_count = 0;
    my @storage_pages;

    # get entries for each host in the pool
    foreach my $host (@$hosts) {
      my ( $filepath, $label, $host_folder, $interface_folder, $url );

      # skip hosts with missing performance data in RRD
      $filepath = XenServerDataWrapper::get_filepath_rrd_host($host);
      unless ( -f $filepath ) { next; }

      # create menu entry
      $label = XenServerDataWrapper::get_label("host", $host);
      $host_folder = XenServerMenu::create_folder($label);
      $url = XenServerMenu::get_url({ type => "host", host => $host });
      my %host_page = %{ XenServerMenu::create_page($label, $url) };
      push @{$host_folder->{children}}, \%host_page;

      # add storages
      my @storages = @{ XenServerDataWrapper::get_host_interface_list($host, "storage") };
      my $storage_count = scalar @storages;
      $storage_total_count += $storage_count;
      if ( $storage_count > 0 ) {
        my $interface_folder = XenServerMenu::create_folder("Storage");
        $totals_url = XenServerMenu::get_url({ type => "storage-aggr", host => $host });
        my %totals_page = %{ XenServerMenu::create_page("Totals", $totals_url) };
        push @{$interface_folder->{children}}, \%totals_page;

        my $item_folder = XenServerMenu::create_folder("Items");
        foreach my $item (@storages) {
          my $sr_uuid = XenServerDataWrapper::complete_sr_uuid($item, $host);
          my $name_label = $sr_uuid ? XenServerDataWrapper::get_label("sr", $sr_uuid) : "no label";
          $label = "$name_label $item";
          $url = XenServerMenu::get_url({ type => "storage", host => $host, storage => $item });
          my %storage_page = %{ XenServerMenu::create_page($label, $url) };
          push @{$item_folder->{children}}, \%storage_page;
          # add the item to the pool-wide list of storages, for "Storage" folder at pool level
          push @storage_pages, \%storage_page;
        }
        push @{$interface_folder->{children}}, $item_folder;

        push @{$host_folder->{children}}, $interface_folder;
      }

      my @networks = @{ XenServerDataWrapper::get_host_interface_list($host, "lan") };
      my $networks_count = scalar @networks;
      if ( $networks_count > 0 ) {
        my $interface_folder = XenServerMenu::create_folder("LAN");
        $totals_url = XenServerMenu::get_url({ type => "net-aggr", host => $host });
        my %totals_page = %{ XenServerMenu::create_page("Totals", $totals_url) };
        push @{$interface_folder->{children}}, \%totals_page;

        my $item_folder = XenServerMenu::create_folder("Items");
        foreach my $item (@networks) {
          $label = $item;
          $url = XenServerMenu::get_url({ type => "net", host => $host, net => $item });
          my %storage_page = %{ XenServerMenu::create_page($label, $url) };
          push @{$item_folder->{children}}, \%storage_page;
        }
        push @{$interface_folder->{children}}, $item_folder;

        push @{$host_folder->{children}}, $interface_folder;
      }

      push @host_entries, $host_folder;
    }

    # add VMs that reside on hosts in this pool
    my $vms;
    if ( -d "$wrkdir/XEN_VMs" ) {
      my @vms = @{ XenServerDataWrapper::get_vm_in_pool_list($pool) };
      my $vm_count = scalar @vms;
      unless ( $vm_count > 0 ) { last; }

      my @pages;
      my ( $label, $url );
      foreach my $uuid (@vms) {
        my $filepath = XenServerDataWrapper::get_filepath_rrd_vm($uuid);
        unless ( -f $filepath ) { next; }

        $label = XenServerDataWrapper::get_label("vm", $uuid);

        $url = XenServerMenu::get_url({ type => "vm", vm => $uuid });
        my %page = %{ XenServerMenu::create_page($label, $url) };
        push @pages, \%page;
      }

      $vms = XenServerMenu::create_folder("VM");
      $vms->{children} = \@pages;
    }

    # add Storages that are assigned to hosts in this pool (the same as under hosts, but here they are in one place)
    my $pool_storages;
    if ( $storage_total_count > 0 ) {
      $pool_storages = XenServerMenu::create_folder("Storage");
      $totals_url = XenServerMenu::get_url({ type => "pool-storage-aggr", pool => $pool });
      my %totals_page = %{ XenServerMenu::create_page("Totals", $totals_url) };
      push @{$pool_storages->{children}}, \%totals_page;

      # add pages collected earlier (under individual hosts)
      my $item_folder = XenServerMenu::create_folder("Items");
      $item_folder->{children} = \@storage_pages;

      push @{$pool_storages->{children}}, $item_folder;
    }

    # add the pool to menu tree
    $label = XenServerDataWrapper::get_label("pool", $pool);
    $pool_folder = XenServerMenu::create_folder($label);
    $totals_url = XenServerMenu::get_url({ type => "pool-aggr", pool => $pool });
    my %totals_page = %{ XenServerMenu::create_page("Totals", $totals_url) };
    unshift @host_entries, \%totals_page;
    $pool_folder->{children} = \@host_entries;
    if ( $vms ) {
      push @{$pool_folder->{children}}, $vms;
    }
    if ( $pool_storages ) {
      push @{$pool_folder->{children}}, $pool_storages;
    }
    push @{$menu_tree->{children}}, $pool_folder;
  }
}

# add "orphaned" VMs that aren't actively reported by any host (could be inactive or removed from current configuration)
# motivation: hypertext links don't work in LPAR2RRD GUI, unless a corresponding item exists in the menu
# WHY IS THIS COMMENTED OUT? because it might be confusing without human-readable labels
# it might come back, in case it is necessary to make sense of historical graphs
#if ( -d "$wrkdir/XEN_VMs" ) {
#  my @vms = @{ XenServerDataWrapper::get_vm_list() };
#  my $host_vm = XenServerDataWrapper::get_conf_section("arch-host-vm");
#
#  if ($host_vm) {
#    while ( my ($host, $host_vms) = each %{$host_vm} ) {
#      foreach my $vm (@$host_vms) {
#        @vms = grep ! /$vm/, @vms;
#      }
#    }
#  }
#
#  my @pages;
#  my ( $label, $url );
#  foreach my $uuid (@vms) {
#    $label = "no label " . substr($uuid, 0, 8) . "...";
#    $url = XenServerMenu::get_url({ type => "vm", vm => $uuid });
#    my %page = %{ XenServerMenu::create_page($label, $url) };
#    push @pages, \%page;
#  }
#
#  if (@pages) {
#    my $vms_folder = XenServerMenu::create_folder("Inactive VMs");
#    $vms_folder->{children} = \@pages;
#    push @{$menu_tree->{children}}, $vms_folder;
#  }
#}

# if there are any nodes, print as JSON
if ($host_exists) {
  my $json = JSON->new->utf8->pretty;
  my $json_data = $json->encode( $menu_tree );
  print $json_data;
}
