# xen-xapi2json.pl
# download XAPI data dump from XenServer hosts

use 5.008_008;

use strict;
use warnings;

use Data::Dumper;

use HostCfg;
use Xorux_lib qw(error);

use LWP::UserAgent;
use HTTP::Request;

use XML::Simple;
use JSON qw(decode_json encode_json);

use POSIX ":sys_wait_h";

defined $ENV{INPUTDIR} || Xorux_lib::error( " INPUTDIR undefined, probably have not read etc/lpar2rrd.cfg " . __FILE__ . ":" . __LINE__ ) && exit 1;

my $inputdir    = $ENV{INPUTDIR};

my $path_prefix         = "$inputdir/data/XEN_iostats";
my $xml_path            = "$path_prefix/xml";
my $json_path           = "$path_prefix/json";
my $metadata_path       = "$path_prefix/metadata";

my $SSH      = $ENV{SSH} . " ";

# create directories in data/
unless ( -d $path_prefix ) {
  mkdir( "$path_prefix", 0755 ) || Xorux_lib::error( " Cannot mkdir $path_prefix: $!" . __FILE__ . ":" . __LINE__ );
}
unless ( -d $xml_path ) {
  mkdir( "$xml_path", 0755 ) || Xorux_lib::error( " Cannot mkdir $xml_path: $!" . __FILE__ . ":" . __LINE__ );
}
unless ( -d $json_path ) {
  mkdir( "$json_path", 0755 ) || Xorux_lib::error( " Cannot mkdir $json_path: $!" . __FILE__ . ":" . __LINE__ );
}
unless ( -d $metadata_path ) {
  mkdir( "$metadata_path", 0755 ) || Xorux_lib::error( " Cannot mkdir $metadata_path: $!" . __FILE__ . ":" . __LINE__ );
}

my %hosts = %{ HostCfg::getHostConnections("XenServer") };
my @pids;
my $pid;
my $timeout = 900;

foreach my $host (keys %hosts) {
  # fork for each host
  unless ( defined( $pid = fork() ) ) {
    Xorux_lib::error( "Error: failed to fork for $host.\n" );
    next;
  } else {
    if ($pid) {
      push @pids, $pid;
    } else {
      local $SIG{ALRM} = sub { die "XenServer XAPI2JSON: $pid timeouted.\n"; };
      alarm($timeout);

      my ( $hostname, $username ) = ( $hosts{$host}{host}, $hosts{$host}{username} );

      if ( $hosts{$host}{auth_api} ) {
        my $password = $hosts{$host}{password};
        xapi2json( $hostname, $username, $password );
      }

      if ( $hosts{$host}{auth_ssh} ) {
        my $ssh_key = $hosts{$host}{ssh_key_id};
        ssh2json( $hostname, $username, $ssh_key );
      }

      exit;
    }
  }
}

# wait for forked data retrieval
for $pid (@pids) {
  waitpid( $pid, 0 );
}

################################################################################

sub xapi2json {
  my ( $host, $username, $password ) = @_;

  my $rrdfile    = "rrd_updates";
  my $start_time = time() - 20*60; # last 20 minutes
  my $db_url     = "http://$host/$rrdfile?start=$start_time&cf=AVERAGE&host=true";

  my $xml_file    = "$xml_path/$rrdfile\_$host\_$start_time.xml";
  my $json_file   = "$json_path/XEN\_$host\_perf_$start_time.json";

  # download dumped RRD as XML
  my $ua = LWP::UserAgent->new;
  my $req = HTTP::Request->new( GET => $db_url );
  $req->content_type( 'application/xml' );
  $req->header( 'Accept'=>'*/*' );
  $req->authorization_basic( "$username", "$password" );
  my $res = $ua->request($req);
  if ( $res->is_success ) {
    if ( $res->{_content} eq "" ) {
      die "error: empty database dump\n";
    }
    else {
      open( XML_FH, "> $xml_file" ) || die "error: cannot save the received RRD XML file\n";
      print XML_FH "$res->{_content}\n";
      close(XML_FH);
    }
  }
  else {
    die "error: " . $res->status_line . "\n";
  }

  # parse XML
  my $data;
  my $xml_simple = XML::Simple->new( keyattr => [], ForceArray => 1 );

  eval { $data = $xml_simple->XMLin($xml_file); };
  if ($@) {
    warning( "XML parsing error: $@" );
    message( "XML parsing error. Trying to recover XML with xmllint" );

    eval {
      my $linted = `xmllint --recover $xml_file`;
      $data = $xml_simple->XMLin($linted);
    };

    if ($@) {
      Xorux_lib::error( "XML parsing error: " . $@ . __FILE__ . ":" . __LINE__ . " file:" . $xml_file );
      message( "XML parsing error: File: " . $xml_file );
      die "error: invalid XML\n";
    }
  }

  # save as JSON
  my $json_data = encode_json( $data );

  open( JSON_FH, "> $json_file" ) || die "error: cannot save the JSON\n";
  print JSON_FH "$json_data\n";
  close(JSON_FH);

  # clean up the downloaded data dump
  unlink $xml_file;
}

sub ssh2json {
  my ( $host, $username, $ssh_key ) = @_;
  my %dictionary;

  my $ssh_cmd = "$SSH ";
  if ( $ssh_key ) {
    $ssh_cmd .= "-i $ssh_key ";
  }
  $ssh_cmd .= "$username\@$host";

  # get labels (human-readable names for UUIDs)
  my @sources = ( "vm", "host", "pool", "sr", "vdi" );
  my ( $cmd_uuids, $cmd_labels, @uuids, @labels, $short_uuid );
  foreach my $source (@sources) {
    $cmd_uuids  = "xe $source-list params=uuid --minimal";
    $cmd_labels = "xe $source-list params=name-label --minimal";

    my @uuids  = split ',', `$ssh_cmd \"$cmd_uuids\"`;
    my @labels = split ',', `$ssh_cmd \"$cmd_labels\"`;

    for my $i (0 .. $#uuids) {
      chomp $uuids[$i];
      chomp $labels[$i];

      unless ( $labels[$i] ) {
        $short_uuid = substr( $uuids[$i], 0, 8 );
        $labels[$i] = "no label $short_uuid ...";
      }

      $dictionary{labels}{$source}{$uuids[$i]} = $labels[$i];
    }
  }

  # get system specification/configuration
  {
    my $cmd_hosts     = "xe host-list params=uuid --minimal";
    my $cmd_vms       = "xe vm-list params=uuid --minimal";

    my @hosts = split ',', `$ssh_cmd \"$cmd_hosts\"`;
    chomp $_ foreach (@hosts);

    my @vms = split ',', `$ssh_cmd \"$cmd_vms\"`;
    chomp $_ foreach (@vms);

    my %host_params = ( address => 'address', memory => 'memory-total',
                        cpu_count => 'cpu_info param-key=cpu_count', socket_count => 'cpu_info param-key=socket_count', cpu_model => 'cpu_info param-key=modelname',
                        version_xen => 'software-version param-key=xen');
    foreach my $host (@hosts) {
      foreach my $key (keys %host_params) {
        my $cmd    = "xe host-param-get param-name=" . $host_params{$key} . " uuid=$host";
        my $result = `$ssh_cmd \"$cmd\" 2>&1`;
        chomp $result;
        if ( $result =~ m/Error: Key name not found in map/ ) { next; }
        if ( $key =~ m/memory/ ) {
          $result /= 1024**3; # convert from bytes to GB
          $result = sprintf "%.2f", $result;
        }
        $dictionary{specification}{host}{$host}{$key} = $result;
      }
    }

    my %vm_params = ( cpu_count => 'VCPUs-number', cpu_count_start => 'VCPUs-at-startup', cpu_count_max => 'VCPUs-max', memory => 'memory-actual', os => 'os-version param-key=name' );
    foreach my $vm (@vms) {
      foreach my $key (keys %vm_params) {
        my $cmd    = "xe vm-param-get param-name=" . $vm_params{$key} . " uuid=$vm";
        my $result = `$ssh_cmd \"$cmd\" 2>&1`;
        chomp $result;
        if ( $result =~ m/Error: Key name not found in map/ ) { next; }
        if ( $key =~ m/memory/ ) {
          $result /= 1024**3; # convert from bytes to GB
          $result = sprintf "%.2f", $result;
        }
        $dictionary{specification}{vm}{$vm}{$key} = $result;
      }
    }

    # TODO add SR (virtual-allocation, physical-utilisation, physical-size) and VDI (virtual-size, physical-utilisation)
    my $cmd_srs     = "xe sr-list params=uuid --minimal";
    my $cmd_vdis    = "xe vdi-list params=uuid --minimal";

    my @srs  = split ',', `$ssh_cmd \"$cmd_srs\"`;
    chomp $_ foreach (@srs);

    my @vdis = split ',', `$ssh_cmd \"$cmd_vdis\"`;
    chomp $_ foreach (@vdis);

    my %sr_params = ( type => 'type', virtual_allocation => 'virtual-allocation', physical_utilisation => 'physical-utilisation', physical_size => 'physical-size' );
    foreach my $sr (@srs) {
      foreach my $key (keys %sr_params) {
        my $cmd    = "xe sr-param-get param-name=" . $sr_params{$key} . " uuid=$sr";
        my $result = `$ssh_cmd \"$cmd\" 2>&1`;
        chomp $result;
        if ( $result =~ m/Error: Key name not found in map/ ) { next; }
        # convert units: if adding other params in the future, skip the calculation
        unless ( $key eq 'type' ) {
          if ( $result == -1 ) { next; }
          $result /= 1000**3; # convert from bytes to GB
          $result = sprintf "%.2f", $result;
        }
        $dictionary{specification}{sr}{$sr}{$key} = $result;
      }
    }

    my %vdi_params = ( virtual_size => 'virtual-size', physical_utilisation => 'physical-utilisation' );
    foreach my $vdi (@vdis) {
      foreach my $key (keys %vdi_params) {
        my $cmd    = "xe vdi-param-get param-name=" . $vdi_params{$key} . " uuid=$vdi";
        my $result = `$ssh_cmd \"$cmd\" 2>&1`;
        chomp $result;
        if ( $result =~ m/Error: Key name not found in map/ ) { next; }
        # convert units: if adding other params in the future, skip the calculation
        if ( $result == -1 ) { next; }
        $result /= 1000**3; # convert from bytes to GB
        $result = sprintf "%.2f", $result;
        $dictionary{specification}{vdi}{$vdi}{$key} = $result;
      }
    }
  }

  # get relations between pools, hosts and VMs
  {
    my $cmd_pool   = "xe pool-list params=uuid --minimal";
    my $cmd_hosts  = "xe host-list params=uuid --minimal";

    my @pools = split ',', `$ssh_cmd \"$cmd_pool\"`;
    my @hosts = split ',', `$ssh_cmd \"$cmd_hosts\"`;

    chomp $_ foreach (@pools);
    chomp $_ foreach (@hosts);

    if ( scalar @pools > 0 ) {
      $dictionary{architecture}{pool}{$pools[0]} = \@hosts;
    }

    foreach my $host (@hosts) {
      my $cmd_vm = "xe vm-list resident-on=$host params=uuid --minimal";
      my @vms    = split ',', `$ssh_cmd \"$cmd_vm\"`;
      chomp $_ foreach (@vms);
      $dictionary{architecture}{host_vm}{$host} = \@vms;
    }
  }

  # get storage relations: (0) pbd=physical device on host (1) sr=LVM volume group, (2) vdi=LVM logical volume, (3) vbd=VM's virtual disk
  # currently, the idea is to present the structure as follows: | sr (host) | vm (vdi) |
  # save relations: (0) sr-hosts (1) vdi-sr (2) vm-vdi
  # i.e., skip pbd and vbd layers
  {
    # get storages (sr)
    my $cmd_sr = "xe sr-list params=uuid --minimal";
    my @sr_list = split ',', `$ssh_cmd \"$cmd_sr\"`;
    chomp $_ foreach (@sr_list);
    foreach my $sr (@sr_list) {
      my $cmd_host = "xe pbd-list params=host-uuid sr-uuid=$sr --minimal";
      my $sr_host_output = `$ssh_cmd \"$cmd_host\"`;
      my @sr_host = split ',', $sr_host_output;
      chomp $_ foreach (@sr_host);
      $dictionary{architecture}{storage}{sr_host}{$sr} = \@sr_host;

      # note that the listing excludes snapshots
      my $cmd_vdi = "xe vdi-list params=uuid sr-uuid=$sr is-a-snapshot=false --minimal";
      my $vdi_cmd_output = `$ssh_cmd \"$cmd_vdi\"`;
      if ( $vdi_cmd_output eq "" ) { next; }
      my @sr_vdi_tmp = split ',', $vdi_cmd_output;
      my @sr_vdi;
      foreach my $vdi (@sr_vdi_tmp) {
        chomp $vdi;
        if ($vdi) {
          push @sr_vdi, $vdi;
        }
      }
      if (@sr_vdi) {
        $dictionary{architecture}{storage}{sr_vdi}{$sr} = \@sr_vdi;
      }
    }
    # vdi-vm
    my $cmd_vbd_vdi = "xe vbd-list params=vdi-uuid --minimal";
    my @vdi_list = split ',', `$ssh_cmd \"$cmd_vbd_vdi\"`;
    chomp $_ foreach (@vdi_list);
    foreach my $vdi (@vdi_list) {
      if ( $vdi eq '<not in database>' ) { next; }
      my $cmd_vm = "xe vbd-list params=vm-uuid vdi-uuid=$vdi --minimal";
      my @vdi_vm = split ',', `$ssh_cmd \"$cmd_vm\"`;
      chomp $_ foreach (@vdi_vm);
      $dictionary{architecture}{storage}{vdi_vm}{$vdi} = \@vdi_vm;
    }
  }

  # save as JSON
  my $labels_file  = "$metadata_path/xe\_$host.json";
  my $json_data;

  $json_data = encode_json( \%dictionary );
  open( JSON_FH, "> $labels_file" ) || Xorux_lib::error( "cannot save the labels JSON file" );
  print JSON_FH "$json_data\n";
  close(JSON_FH);
}
